package usboszilloskop.model.analyze;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * analyze a mathematical expression
 *
 * @author eulerscheZahl
 */
public class Parser {

    private class Node {

        public Node right, left;
        public String op;

        public Node() {
            this.op = "";
        }

        @Override
        public String toString() {
            if (this.right == null && this.left == null) //only a constant
            {
                return this.op;
            }
            if (this.left == null) //e.g. sin(x)
            {
                if (this.op.equals("-") && getPriority(this.right.op) > getPriority(this.op)) {
                    return this.op + this.right.toString();
                }
                return this.op + "(" + this.right.toString() + ")";
            }
            //operator with right and left value
            String l = this.left.toString();
            String r = this.right.toString();
            if (getPriority(this.op) > getPriority(this.left.op)) {
                l = "(" + l + ")";
            }
            if (getPriority(this.op) > getPriority(this.right.op) || (this.op.equals("-") && this.right.op.equals("+") || this.op.equals("/") && this.right.op.equals("*"))) {
                r = "(" + r + ")";
            }
            return (l + this.op + r);
        }

        private boolean IsNumber() {
            boolean result;
            try {
                Double.parseDouble(this.op);
                result = true;
            } catch (NumberFormatException e) {
                result = false;
            }
            return result;
        }

        private void Null() {
            this.right = null;
            this.left = null;
        }

        public void Zuweisung(Node k) {
            this.op = k.op;
            this.right = k.right;
            this.left = k.left;
        }
    }

    private static class Operator {

        public String op;
        public int prioritaet;

        public Operator(String op, int prioritaet) {
            this.prioritaet = prioritaet;
            this.op = op;
        }
    }

    static final Operator[] operators = {
        new Operator(";", 0), //base for logarithms
        new Operator("+", 1), new Operator("-", 1),
        new Operator("*", 2), new Operator("/", 2),
        new Operator("^", 3),
        new Operator("sin", 4), new Operator("cos", 4), new Operator("tan", 4), new Operator("lg", 4), new Operator("ln", 4), new Operator("log", 4), new Operator("abs", 4), new Operator("sigma", 4)
    };

    private static int getPriority(String operat) {
        for (Operator o : operators) {
            if (operat.equals(o.op)) {
                return o.prioritaet;
            }
        }
        return 10;
    }

    private Node root;

    /**
     * parse a term
     *
     * @param term term to analyze
     */
    public Parser(String term) {
        term = simpleInput(term);
        root = new Node();
        root = parseRecurs(term, root);
    }

    /**
     * get all the vars used in a term
     *
     * @return vars in an array
     */
    public String[] getVars() {
        ArrayList<String> result = new ArrayList<String>();
        getVarsRecurs(result, this.root);
        return result.toArray(new String[result.size()]);
    }

    private void getVarsRecurs(ArrayList<String> vars, Node n) {
        if (n.right == null && n.left == null) {
            try {
                Float.parseFloat(n.op);
            } catch (NumberFormatException e) {
                boolean append = true;
                if (n.op.equalsIgnoreCase("e") || n.op.equalsIgnoreCase("pi")) {
                    append = false;
                }
                for (String s : vars) {
                    if (s.equals(n.op)) {
                        append = false;
                    }
                }
                if (append) {
                    vars.add(n.op);
                }
            }
        } else {
            if (n.left != null) {
                getVarsRecurs(vars, n.left);
            }
            if (n.right != null) {
                getVarsRecurs(vars, n.right);
            }
        }
    }

    /**
     * simplify a term
     */
    public void Simplify() {
        String tmp;
        do {
            tmp = this.toString();
            simplifyRecurs(root);
        } while (!tmp.equals(this.toString()));
    }

    /**
     * calculate the value of a term for given values to use for the variables
     *
     * @param values the values to put into the term
     * @return the calculated value
     */
    public float calc(HashMap<String, Float> values) {
        return calcRecurs(values, this.root);
    }

    private static float calcRecurs(HashMap<String, Float> values, Node n) {
        if (n.right == null && n.left == null) //only a constant remaining
        {
            if (values.containsKey(n.op)) {
                return values.get(n.op); //insert value for variable
            }
            if (n.op.equals("e")) {
                return (float) Math.E; //allow overriding of pi and e
            }
            if (n.op.toLowerCase().equals("pi")) {
                return (float) Math.PI;
            }
            try {
                return Float.parseFloat(n.op);
            } catch (NumberFormatException e) {
                return 0;
            }
        }
        if (n.left == null) //e.g. sin(x)
        {
            switch (n.op) {
                case "-":
                    return -calcRecurs(values, n.right); //unary minus
                case "sin":
                    return (float) Math.sin(calcRecurs(values, n.right));
                case "cos":
                    return (float) Math.cos(calcRecurs(values, n.right));
                case "tan":
                    return (float) Math.tan(calcRecurs(values, n.right));
                case "ln":
                    return (float) Math.log(calcRecurs(values, n.right));
                case "lg":
                    return (float) Math.log10(calcRecurs(values, n.right));
                case "log":
                    return (float) Math.log(calcRecurs(values, n.right.right)) / (float) Math.log(calcRecurs(values, n.right.left));
                case "abs":
                    return (float) Math.abs(calcRecurs(values, n.right));
                case "sigma":
                    return calcRecurs(values, n.right) < 0 ? 0 : 1;
            }
        }
        //operator with right and left value
        switch (n.op) {
            case "+":
                return calcRecurs(values, n.left) + calcRecurs(values, n.right);
            case "-":
                return calcRecurs(values, n.left) - calcRecurs(values, n.right);
            case "*":
                return calcRecurs(values, n.left) * calcRecurs(values, n.right);
            case "/":
                return calcRecurs(values, n.left) / calcRecurs(values, n.right);
            case "^":
                return (float) Math.pow(calcRecurs(values, n.left), calcRecurs(values, n.right));
        }
        return 0;
    }

    /**
     * derive a function
     *
     * @param d the parameter the function is derived to
     * @return a new Parser holding the term of the derived function
     */
    public Parser derive(String d) {
        Parser result = new Parser(deriveRecurs(d, this.root));
        result.Simplify();
        return result;
    }

    private String deriveRecurs(String d, Node n) {
        if (n.right == null && n.left == null) //only a constant
        {
            if (n.op.equals(d)) {
                return "1";
            }
            return "0";
        }
        if (n.left == null) //e.g. sin(x)
        {
            switch (n.op) {
                case "-":
                    return "(-(" + deriveRecurs(d, n.right) + "))";
                case "sin":
                    return "(cos(" + n.right.toString() + ")*(" + deriveRecurs(d, n.right) + "))";
                case "cos":
                    return "(-sin(" + n.right.toString() + ")*(" + deriveRecurs(d, n.right) + "))";
                case "tan":
                    return "(1/cos(" + n.right.toString() + ")^2*(" + deriveRecurs(d, n.right) + "))";
                case "ln":
                    return "(1/(" + n.right.toString() + ")*(" + deriveRecurs(d, n.right) + "))";
                case "lg":
                    return "(1/(ln(10)*(" + n.right.toString() + ")*(" + deriveRecurs(d, n.right) + ")))";
                case "log":
                    return "(1/(ln(" + n.right.left + ")*(" + n.right.right.toString() + ")*(" + deriveRecurs(d, n.right.right) + ")))";
                case "abs":
                    return "((" + n.right.toString() + ")/abs(" + n.right.toString() + ")*(" + deriveRecurs(d, n.right) + "))";
                case "sigma":
                    return "0";
            }
        }
        switch (n.op) {
            case "+":
                return "((" + deriveRecurs(d, n.left) + ")+(" + deriveRecurs(d, n.right) + "))";
            case "-":
                return "((" + deriveRecurs(d, n.left) + ")-(" + deriveRecurs(d, n.right) + "))";
            case "*":
                return "(" + deriveRecurs(d, n.left) + ")*(" + n.right.toString() + ")+(" + n.left.toString() + ")*(" + deriveRecurs(d, n.right) + ")";
            case "/":
                return "((" + n.right.toString() + ")*(" + deriveRecurs(d, n.left) + ")-(" + n.left.toString() + ")*(" + deriveRecurs(d, n.right) + "))/((" + n.right.toString() + ")^2)";
            case "^":
                if (n.left.op.equals("e")) {
                    return "((" + n.toString() + ")*(" + deriveRecurs(d, n.right) + "))"; //derive exp(...)
                }
                return "(((" + n.left.toString() + ")^((" + n.right.toString() + ")-1))*(((" + n.right.toString() + ")*(" + deriveRecurs(d, n.left) + "))+((" + n.left.toString() + ")*ln(" + n.left.toString() + ")*(" + deriveRecurs(d, n.right) + "))))";
        }
        return "0";
    }

    private void simplifyRecurs(Node k) {
        if (k.left == null && k.right == null) {
            return;
        }
        if (k.left != null) {
            simplifyRecurs(k.left);
        }
        if (k.right != null) {
            simplifyRecurs(k.right);
        }
        if (k.left == null) {
            //ist eine Funktion
            switch (k.op) {
                case "abs":
                    if (k.right.IsNumber()) {
                        k.op = String.valueOf(Math.abs(Double.parseDouble(k.right.op)));
                        k.Null();
                        return;
                    }
                    return;
                case "-":
                    if (k.op.equals("-") && k.right != null && k.right.op.equals("-") && k.right.left == null) //  -(-(x)) = x
                    {
                        k.Zuweisung(k.right.right);
                        return;
                    }
                    return;
            }
            return;
        }
        if (k.left != null && k.left != null) {
            boolean numL = k.left.IsNumber();
            boolean numR = k.right.IsNumber();
            switch (k.op) {
                case "+":
                case "-":
                    if (numL && k.left.op.equals("0")) {
                        if (k.op.equals("+")) {
                            k.Zuweisung(k.right);
                        } else {
                            k.left = null;
                        }
                        return;
                    }
                    if (numR && k.right.op.equals("0")) {
                        k.Zuweisung(k.left);
                        return;
                    }
                    if (numR && numL) {
                        if (k.op.equals("+")) {
                            k.op = String.valueOf(Double.parseDouble(k.left.op) + Double.parseDouble(k.right.op));
                        }
                        if (k.op.equals("-")) {
                            k.op = String.valueOf(Double.parseDouble(k.left.op) - Double.parseDouble(k.right.op));
                        }
                        k.Null();
                        return;
                    }
                    if (k.op.equals("+") && k.right.op.equals("-") && k.right.left == null) //  x+(-y) = x-y
                    {
                        k.right.Zuweisung(k.right.right);
                        k.op = "-";
                        return;
                    }
                    if (k.op.equals("-") && k.right.op.equals("-") && k.right.left == null) //  x-(-y) = x+y
                    {
                        k.right.Zuweisung(k.right.right);
                        k.op = "+";
                        return;
                    }
                    return;
                case "*":
                    if (numL && k.left.op.equals("1")) {
                        k.Zuweisung(k.right);
                        return;
                    }
                    if (numR && k.right.op.equals("1")) {
                        k.Zuweisung(k.left);
                        return;
                    }
                    if (numL && k.left.op.equals("0") || numR && k.right.op.equals("0")) {
                        k.op = "0";
                        k.Null();
                        return;
                    }
                    if (numR && numL) {
                        k.op = String.valueOf(Double.parseDouble(k.left.op) * Double.parseDouble(k.right.op));
                        k.Null();
                        return;
                    }
                    if (!numL && numR) //flip x*2 to 2*x
                    {
                        Node tmp = new Node();
                        tmp.Zuweisung(k);
                        k.right = tmp.left;
                        k.left = tmp.right;
                        return;
                    }
                    return;
                case "/":
                    if (numR && numL) {
                        k.op = String.valueOf(Double.parseDouble(k.left.op) / Double.parseDouble(k.right.op));
                        k.Null();
                        return;
                    }
                    return;
                case "^":
                    if (numR && k.right.op.equals("0")) //x^0 = 1
                    {
                        k.op = "1";
                        k.Null();
                        return;
                    }
                    if (numR && k.right.op.equals("1")) {
                        k.Zuweisung(k.left);
                        return;
                    }
                    if (numR && numL) {
                        k.op = String.valueOf(Math.pow(Double.parseDouble(k.left.op), Double.parseDouble(k.right.op)));
                        k.Null();
                        return;
                    }
                    return;
            }
            return;
        }
    }

    /**
     * convert the term to a string
     *
     * @return the function
     */
    @Override
    public String toString() {
        return root.toString();
    }

    private String simpleInput(String term) {
        term = term.replace(" ", "");
        term = term.replace("²", "^2");
        term = term.replace("³", "^3");
        term = term.replace(")(", ")*(");
        while (true) //3x to 3*x; 3sin(x) to 3*sin(x)
        {
            Pattern pattern = Pattern.compile("[0-9][a-zA-Z]");
            Matcher m = pattern.matcher(term);
            if (m.find()) {
                term = term.substring(0, m.start() + 1) + "*" + term.substring(m.start() + 1);
            } else {
                break;
            }
        }
        int betrag = 0;
        for (int i = 0; i < term.length(); i++) {
            if (term.charAt(i) == '|') {
                if (++betrag % 2 == 1) {
                    term = term.substring(0, i - 1) + "abs(" + term.substring(i + 1);
                } else {
                    term = term.substring(0, i - 1) + ")" + term.substring(i + 1);
                }
            }
        }
        return term;
    }

    private Node parseRecurs(String term, Node n) {
        String[] s = parse(term);
        if (s[1].equals("")) //not a function, only an argument
        {
            n.op = s[0];
            n.Null();
        } else if (s[0].equals("")) //e.g. sin(x)
        {
            n.op = s[1];
            n.right = new Node();
            n.right = parseRecurs(s[2], n.right);
            if (n.op.equals("+")) {
                n.Zuweisung(n.right); //delete unary plus
            }
            n.left = null;
        } else //operator with right and left value
        {
            n.op = s[1];
            n.right = new Node();
            n.left = new Node();
            n.left = parseRecurs(s[0], n.left);
            n.right = parseRecurs(s[2], n.right);
        }
        return n;
    }

    private String[] parse(String term) {
        //find operators with lowest priority - outside of brackets
        //go from left to right
        String[] result = new String[3];
        int brackets = 0, currPrio = 10, position = term.length();
        String currOp = "";
        for (int i = term.length() - 1; i >= 0; i--) {
            if (term.charAt(i) == ')') {
                brackets++;
            }
            if (term.charAt(i) == '(') {
                brackets--;
            }
            if (brackets != 0) {
                continue;
            }
            for (Operator op : operators) {
                if (op.prioritaet < currPrio && i + op.op.length() < term.length() && op.op.equals(term.substring(i, i + op.op.length()))) {
                    currPrio = op.prioritaet;
                    position = i;
                    currOp = op.op;
                }
            }
        }
        if (currPrio == 10 && term.startsWith("(")) {
            return parse(term.substring(1, term.length() - 1));
        }
        result[0] = term.substring(0, position);
        result[1] = currOp;
        result[2] = term.substring(position + currOp.length());
        return result;
    }
}
